#include <cstddef> 
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <iterator>
#include <ostream>
#include <string>
#include <vector>

#include <boost/filesystem.hpp>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <GL/glew.h>
#include <GL/glx.h>


//-------------------------------------------------------------------------------------------------
// Read ascii file, return as string
//

std::string read_ascii(std::string const& filename)
{
    std::ifstream file(filename.c_str());

    if (!file.is_open())
    {
        return "";
    }

    return std::string(
        std::istreambuf_iterator<char>(file.rdbuf()),
        std::istreambuf_iterator<char>());
}

 
//-------------------------------------------------------------------------------------------------
// Try to load vertex and fragment shaders
//

bool try_load(
        std::string vert_path,
        std::string frag_path,
        std::string& vert_source,
        std::string& frag_source
        )
{
    namespace fs = boost::filesystem;

    std::cout << "Try to load shaders from:\n";
    std::cout << "  " << vert_path << '\n';
    std::cout << "  " << frag_path << '\n';
    if (fs::exists(vert_path) && fs::exists(frag_path))
    {
        vert_source = read_ascii(vert_path);
        frag_source = read_ascii(frag_path);
        std::cout << "Success!\n";
        return true;
    }
    else
    {
        std::cout << "Files do not exist..\n";
        return false;
    }
}


//-------------------------------------------------------------------------------------------------
// Load shaders in shaders/ folder
//

bool load_shaders(std::string& vert_source, std::string& frag_source)
{
    std::string vert_filename = "vertex_shader.vert";
    std::string frag_filename = "fragment_shader.frag";

    // Try to find the two shader files
    bool loaded = false;

    // Environment variable SHADER_DIR overrides!
    char* shader_dir = getenv("SHADER_DIR");
    if (!loaded && shader_dir)
    {
        loaded = try_load(
            std::string(shader_dir) + '/' + vert_filename,
            std::string(shader_dir) + '.' + frag_filename,
            vert_source,
            frag_source
            );
    }

    if (!loaded)
    {
        loaded = try_load("shaders/" + vert_filename, "shaders/" + frag_filename, vert_source, frag_source);
    }

    if (!loaded)
    {
        loaded = try_load(vert_filename, frag_filename, vert_source, frag_source);
    }

    return loaded;
}


//-------------------------------------------------------------------------------------------------
// Compile the shader program
//

bool compile_shaders(std::string vert_source, std::string frag_source)
{
    GLchar const* vert_source_c = vert_source.c_str();
    GLchar const* frag_source_c = frag_source.c_str();
    GLint vert_len = static_cast<GLint>(vert_source.length());
    GLint frag_len = static_cast<GLint>(frag_source.length());

    GLuint vs = glCreateShader(GL_VERTEX_SHADER);
    GLuint fs = glCreateShader(GL_FRAGMENT_SHADER);

    glShaderSource(
        vs,
        1,
        &vert_source_c,
        &vert_len
        );

    glShaderSource(
        fs,
        1,
        &frag_source_c,
        &frag_len
        );

    // Compile the two shaders here

    // Then check success, print compiler errors otherwise
    GLint success = GL_FALSE;


    // Link the two shaders to a single program

    // Eventually return if compilation was successful

    return success;
}


//-------------------------------------------------------------------------------------------------
// Main function, creates a window with an OpenGL context and tries to compile the shader program
//

int main (int argc, char** argv)
{
    Display *display = XOpenDisplay(0);
 
    if (!display)
    {
        std::cerr <<  "Failed to open X display\n";
        exit(EXIT_FAILURE);
    }
 
    // Request RGBA8, DEPTH24 framebuffer with
    // double buffering
    static int visual_attribs[] = {
        GLX_X_RENDERABLE    , True,
        GLX_DRAWABLE_TYPE   , GLX_WINDOW_BIT,
        GLX_RENDER_TYPE     , GLX_RGBA_BIT,
        GLX_X_VISUAL_TYPE   , GLX_TRUE_COLOR,
        GLX_RED_SIZE        , 8,
        GLX_GREEN_SIZE      , 8,
        GLX_BLUE_SIZE       , 8,
        GLX_ALPHA_SIZE      , 8,
        GLX_DEPTH_SIZE      , 24,
        GLX_STENCIL_SIZE    , 8,
        GLX_DOUBLEBUFFER    , True,
        GLX_SAMPLE_BUFFERS  , 1,
        GLX_SAMPLES         , 4,
        None
        };
 
    int fbcount = 0;
    GLXFBConfig* fbconfig = glXChooseFBConfig(
        display,
        DefaultScreen(display),
        visual_attribs,
        &fbcount
        );

    if (!fbconfig || fbcount < 1)
    {
        std::cerr << "Failed to choose framebuffer config\n";
        exit(EXIT_FAILURE);
    }

    // Get a visual for the framebuffer config 
    XVisualInfo *vi = glXGetVisualFromFBConfig(display, fbconfig[0]/*TODO: choose best match*/);

    // Create colormap 
    XSetWindowAttributes swa;
    swa.colormap = XCreateColormap(
        display,
        RootWindow(display, vi->screen),
        vi->visual, AllocNone
        );
    swa.background_pixmap = None ;
    swa.border_pixel = 0;
    swa.event_mask = StructureNotifyMask;

    // Create a 1x1 pixel window 
    Window win = XCreateWindow(
        display,
        RootWindow(display, vi->screen), 
        0, 0, 1, 1, 0,
        vi->depth,
        InputOutput, 
        vi->visual, 
        CWBorderPixel | CWColormap | CWEventMask,
        &swa
        );

    if (!win)
    {
        std::cerr << "Failed to create window\n";
        exit(EXIT_FAILURE);
    }
 
    XMapWindow(display, win);
 
    GLXContext ctx = glXCreateContext(display, vi, 0, True);
 
    XFree(fbconfig);

    // Check that we have a direct rendering context 
    if (!glXIsDirect(display, ctx))
    {
        std::cerr << "Failed to create a direct rendering context\n";
        exit(EXIT_FAILURE);
    }

    // Make the context current so we can do OpenGL things 
    if (!glXMakeCurrent(display, win, ctx))
    {
        std::cerr << "Cannot make context current\n";
        exit(EXIT_FAILURE);
    }


    // OpenGL code goes here..

    // GLEW makes extension loading way simpler
    // For simplicity, we create a legacy OpenGL context where
    // functions like glCreateShader or glCompileShader
    // are still extensions..
    glewInit();

    std::string vert_source = "";
    std::string frag_source = "";

    if (load_shaders(vert_source, frag_source))
    {
        if (!compile_shaders(vert_source, frag_source))
        {
            std::cerr << "Failed to compile shaders\n";
        }
    }
    else
    {
        std::cerr << "Failed to load shader source\n";
    }


    // Cleanup
    ctx = glXGetCurrentContext();
    glXMakeCurrent(display, 0, 0);
    glXDestroyContext(display, ctx);
    XDestroyWindow(display, win);
}
